;;;; uuid-lib.scm
;;;; Kon Lovett, May '06
;;;; BSD

;;; Wrapper for Theodore Y. Ts'o OSF DCE 1.1 UUID Suite
;;; http://lib.sourceforge.net/
;;; Part of MacOS X & Linux (every modern unix?)

(declare
  (bound-to-procedure
    ##sys#signal-hook))

(module uuid-lib

(;export
  ;
  uuid?
  uuid-null?
  uuid-compare
  uuid=? uuid<? uuid>? uuid<=? uuid>=?
  uuid-copy
  uuid-clear!
  make-uuid
  string->uuid uuid->string
  ;
  uuid-generate
  uuid-parse
  uuid-unparse)

(import scheme)
(import (chicken base))
(import (chicken type))
(import (chicken condition))
(import (chicken foreign))
(import (chicken memory))
(import (chicken gc))
(import (chicken locative))

(define-type boxed-uuid pointer)

(: uuid? (* -> boolean : boxed-uuid))
(: check-uuid (symbol * -> boxed-uuid))
(: uuid-null? (* --> boolean))
(: uuid-compare (boxed-uuid boxed-uuid --> fixnum))
(: uuid=? (boxed-uuid boxed-uuid --> boolean))
(: uuid<? (boxed-uuid boxed-uuid --> boolean))
(: uuid>? (boxed-uuid boxed-uuid --> boolean))
(: uuid<=? (boxed-uuid boxed-uuid --> boolean))
(: uuid>=? (boxed-uuid boxed-uuid --> boolean))
(: uuid-clear! (boxed-uuid -> boxed-uuid))
(: uuid-copy (boxed-uuid --> boxed-uuid))
(: uuid-generate (#!optional (or false symbol) --> boxed-uuid))
(: make-uuid (#!optional (or false symbol) --> boxed-uuid))
(: uuid-parse ((or string symbol) --> boxed-uuid))
(: uuid-unparse (boxed-uuid #!optional (or false symbol) --> string))
(: string->uuid (string --> boxed-uuid))
(: uuid->string (boxed-uuid --> string))

;;

#>
#include <uuid/uuid.h>
<#

(define UUID-SIZEOF (foreign-value "sizeof(uuid_t)" int))
(define-constant UUID-STRING-LENGTH 36)

(define uuid_is_null
  (foreign-lambda bool uuid_is_null nonnull-c-pointer))

(define uuid_compare
  (foreign-lambda int uuid_compare nonnull-c-pointer nonnull-c-pointer))

(define uuid_clear
  (foreign-lambda void uuid_clear nonnull-c-pointer))

(define uuid_copy
  (foreign-lambda void uuid_copy nonnull-c-pointer nonnull-c-pointer))

(define uuid_generate
  (foreign-lambda void uuid_generate nonnull-c-pointer))

(define uuid_generate_random
  (foreign-lambda void uuid_generate_random nonnull-c-pointer))

(define uuid_generate_time
  (foreign-lambda void uuid_generate_time nonnull-c-pointer))

(define uuid_parse
  (foreign-lambda int uuid_parse c-string nonnull-c-pointer))

(define uuid_unparse
  (foreign-lambda void uuid_unparse nonnull-c-pointer nonnull-c-pointer))

(define uuid_unparse_upper
  (foreign-lambda void uuid_unparse_upper nonnull-c-pointer nonnull-c-pointer))

(define uuid_unparse_lower
  (foreign-lambda void uuid_unparse_lower nonnull-c-pointer nonnull-c-pointer))

;;

(define-record-type uuid
  (%make-uuid puuid)
  %uuid?
  (puuid %uuid-puuid))

(define-inline (unbox-puuid boxed-puuid)
  (%uuid-puuid boxed-puuid) )

(define-inline (boxed-puuid-compare uuid1 uuid2)
  (uuid_compare (unbox-puuid uuid1) (unbox-puuid uuid2)) )

(define (free-uuid uuid)
  (free (unbox-puuid uuid)) )

(define-inline (box-puuid puuid)
  (set-finalizer! (%make-uuid puuid) free-uuid) )

(define-inline (new-uuid)
  (box-puuid (allocate UUID-SIZEOF)) )

;;

(define (uuid? obj)
  (%uuid? obj) )

(define (check-uuid loc obj)
  (unless (%uuid? obj)
    (##sys#signal-hook #:type-error 'uuid-compare' "bad argument type - not a uuid" obj))
  obj )

(define (uuid-null? obj)
  (and
    (%uuid? obj)
    (uuid_is_null (unbox-puuid obj))) )

(define (uuid-compare uuid1 uuid2)
  (let ((cmp
          (boxed-puuid-compare
            (check-uuid 'uuid-compare uuid1)
            (check-uuid 'uuid-compare uuid2))) )
    (cond
      ((negative? cmp) -1)
      ((zero? cmp) 0)
      (else 1) ) ) )

(define (uuid=? uuid1 uuid2)
  (zero? (boxed-puuid-compare uuid1 uuid2)) )

(define (uuid<? uuid1 uuid2)
  (negative? (boxed-puuid-compare uuid1 uuid2)) )

(define (uuid>? uuid1 uuid2)
  (positive? (boxed-puuid-compare uuid1 uuid2)) )

(define (uuid<=? uuid1 uuid2)
  (let ((comp (boxed-puuid-compare uuid1 uuid2)))
    (or (zero? comp) (negative? comp))) )

(define (uuid>=? uuid1 uuid2)
  (let ((comp (boxed-puuid-compare uuid1 uuid2)))
    (or (zero? comp) (positive? comp))) )

(define (uuid-clear! uuid)
  (check-uuid 'uuid-clear! uuid)
  (and (uuid_clear (unbox-puuid uuid))
       uuid) )

(define (uuid-copy old-uuid)
  (check-uuid 'uuid-copy old-uuid)
  (let ((uuid (new-uuid)))
    (uuid_copy (unbox-puuid uuid) (unbox-puuid old-uuid))
    uuid) )

(define (uuid-generate #!optional (variant #f))
  (let ((uuid (new-uuid)))
    (case variant
      ((#f)
        (uuid_generate (unbox-puuid uuid)))
      ((V4 random)
        (uuid_generate_random (unbox-puuid uuid)))
      ((V1 time)
        (uuid_generate_time (unbox-puuid uuid)))
      (else
        (##sys#signal-hook #:type-error 'uuid-generate
          "bad argument type - invalid variant" variant)))
    uuid ) )

(define (make-uuid . args)
  (if (null? args)
      (uuid-clear! (new-uuid))
      (apply uuid-generate args) ) )

(define (uuid-parse uuid-text)
  (unless (or (string? uuid-text) (symbol? uuid-text))
    (##sys#signal-hook #:type-error 'uuid-parse
      "bad argument type - not a string or symbol" uuid-text))
  (let* ((uuid-text
           (if (symbol? uuid-text) (symbol->string uuid-text) uuid-text))
         (uuid
           (new-uuid)) )
    (and (zero? (uuid_parse uuid-text (unbox-puuid uuid)))
         uuid) ) )

(define (uuid-unparse uuid #!optional (kase #f))
  (check-uuid 'uuid-unparse uuid)
  (let* ((uuid-text (make-string (add1 UUID-STRING-LENGTH)))
         (puuid-text (make-locative uuid-text)) )
    (case kase
      ((upper)
        (uuid_unparse_upper (unbox-puuid uuid) puuid-text))
      ((lower)
        (uuid_unparse_lower (unbox-puuid uuid) puuid-text))
      ((#f)
        (uuid_unparse (unbox-puuid uuid) puuid-text))
      (else
        (##sys#signal-hook #:type-error 'uuid-generate
          "bad argument type - invalid case" kase)) )
    (substring uuid-text 0 UUID-STRING-LENGTH) ) )

(define (string->uuid str)
  (uuid-parse str) )

(define (uuid->string uuid)
  (uuid-unparse uuid 'lower) )

) ;module uuid-lib
